"""
Module defining classes for Hilbert spaces, linear operators and linear forms. 

The classes within this module have interdependencies and so cannot be split 
into separate modules. 
"""

import numpy as np
from scipy.sparse.linalg import LinearOperator as ScipyLinOp
from scipy.sparse import diags

from pygeoinf.random_matrix import (
    fixed_rank_random_range,
    variable_rank_random_range,
    random_svd,
    random_cholesky,
    random_eig,
)


class HilbertSpace:
    """
    A class for real Hilbert spaces. To define an instance, the
    user needs to provide the following:

        (1) The dimension of the space, or the dimension of the
            finite-dimensional approximating space.
        (2) A mapping from elements of the space to their components.
            These components must be expressed as numpy arrays with
            shape (dim) with dim the space's dimension.
        (3) A mapping from components back to the vectors. This
            needs to be the inverse of the mapping in (2), but
            this requirement is not automatically checked.
        (4) The inner product on the space.
        (5) The mapping from the space to its dual.
        (6) The mapping from a dual vector to its representation
            within the space.

    Optinally, custom implementations for the basic vector operations can
    be provided.
    """

    def __init__(
        self,
        dim,
        to_components,
        from_components,
        inner_product,
        to_dual,
        from_dual,
        /,
        *,
        add=None,
        subtract=None,
        multiply=None,
        axpy=None,
        copy=None,
        vector_multiply=None,
        base=None,
    ):
        """
        Args:
            dim (int): The dimension of the space, or of the
                finite-dimensional approximating space.
            to_components (callable):  A functor that maps vectors
                to their components.
            from_components (callable): A functor that maps components
                to vectors.
            inner_product (callable): A functor the implements the inner
                product on the space.
            to_dual (callable | None): A funcator that maps a vector
                to the cannonically associated dual vector.
            from_dual (callable | None): A functor that maps a dual vector
                to its representation on the space.
            add (callable): Implements vector addition.
            subtract (callable): Implements vector subtraction.
            multiply (callable): Implements scalar multiplication.
            axpy (callable): Implements the mapping y -> a*x + y
            copy (callable): Implements deep copy of a vector.
            base (HilbertSpace | None): Set to none for an original space,
                and to the base space when forming the dual.
        """
        self._dim = dim
        self.__to_components = to_components
        self.__from_components = from_components
        self.__inner_product = inner_product
        self.__from_dual = from_dual
        self.__to_dual = to_dual
        self._base = base
        self._add = self.__add if add is None else add
        self._subtract = self.__subtract if subtract is None else subtract
        self._multiply = self.__multiply if multiply is None else multiply
        self._axpy = self.__axpy if axpy is None else axpy
        self._copy = self.__copy if copy is None else copy
        self._vector_multiply = vector_multiply

    @property
    def dim(self):
        """The dimension of the space."""
        return self._dim

    @property
    def has_vector_multiply(self):
        """
        True if multiplication of elements is defined.
        """
        return self._vector_multiply is not None

    @property
    def dual(self):
        """The dual of the Hilbert space."""
        if self._base is None:
            return HilbertSpace(
                self.dim,
                self._dual_to_components,
                self._dual_from_components,
                self._dual_inner_product,
                self.from_dual,
                self.to_dual,
                base=self,
            )
        else:
            return self._base

    @property
    def zero(self):
        """
        Returns the zero vector for the space.
        """
        return self.from_components(np.zeros((self.dim)))

    @property
    def coordinate_inclusion(self):
        """
        Returns the linear operator that maps coordinate vectors
        to elements of the sapce.
        """
        domain = EuclideanSpace(self.dim)

        def dual_mapping(xp):
            cp = self.dual.to_components(xp)
            return domain.to_dual(cp)

        def adjoint_mapping(y):
            yp = self.to_dual(y)
            return self.dual.to_components(yp)

        return LinearOperator(
            domain,
            self,
            self.from_components,
            dual_mapping=dual_mapping,
            adjoint_mapping=adjoint_mapping,
        )

    @property
    def coordinate_projection(self):
        """
        Returns the linear operator that maps vectors to their coordinates.
        """
        codomain = EuclideanSpace(self.dim)

        def dual_mapping(cp):
            c = codomain.from_dual(cp)
            return self.dual.from_components(c)

        def adjoint_mapping(c):
            xp = self.dual.from_components(c)
            return self.from_dual(xp)

        return LinearOperator(
            self,
            codomain,
            self.to_components,
            dual_mapping=dual_mapping,
            adjoint_mapping=adjoint_mapping,
        )

    @property
    def riesz(self):
        """
        Returns as a LinearOpeator the isomorphism from the
        dual space to the space via the Riesz representation theorem.
        """
        return LinearOperator.self_dual(self.dual, self.from_dual)

    @property
    def inverse_riesz(self):
        """
        Returns as a LinearOpeator the isomorphism from the
        space to the dual space via the Riesz representation theorem.
        """
        return LinearOperator.self_dual(self, self.to_dual)

    def inner_product(self, x1, x2):
        """Return the inner product of two vectors."""
        return self.__inner_product(x1, x2)

    def squared_norm(self, x):
        """Return the squared norm of a vector."""
        return self.inner_product(x, x)

    def norm(self, x):
        """Return the norm of a vector."""
        return np.sqrt(self.squared_norm(x))

    def gram_schmidt(self, vectors):
        """
        Given a list of elements of the space, returns a set
        of orthonormal vectors spanning the same sub-space.
        """

        if not all(self.is_element(vector) for vector in vectors):
            raise ValueError("Not all vectors are elements of the space")

        orthonormalised_vectors = []
        for i, vector in enumerate(vectors):
            for j in range(i):
                product = self.inner_product(vector, orthonormalised_vectors[j])
                vector = self.axpy(-product, orthonormalised_vectors[j], vector)
            norm = self.norm(vector)
            vector = self.multiply(1 / norm, vector)
            orthonormalised_vectors.append(vector)

        return orthonormalised_vectors

    def to_dual(self, x):
        """Map a vector to cannonically associated dual vector."""
        return self.__to_dual(x)

    def from_dual(self, xp):
        """Map a dual vector to its representation in the space."""
        return self.__from_dual(xp)

    def _dual_inner_product(self, xp1, xp2):
        return self.inner_product(self.from_dual(xp1), self.from_dual(xp2))

    def is_element(self, x):
        """
        Returns True if the argument is a vector in the space.
        """
        return isinstance(x, type(self.zero))

    def add(self, x, y):
        """Returns x + y."""
        return self._add(x, y)

    def subtract(self, x, y):
        """Returns x - y."""
        return self._subtract(x, y)

    def multiply(self, a, x):
        """Returns a * x."""
        return self._multiply(a, x)

    def negative(self, x):
        """Returns -x."""
        return self.multiply(-1, x)

    def axpy(self, a, x, y):
        """
        Sets y = y + a * x.
        """
        return self._axpy(a, x, y)

    def copy(self, x):
        """
        Returns a copy of x.
        """
        return self._copy(x)

    def vector_multiply(self, x1, x2):
        """
        Returns the product of two elements of the space.
        """
        if not self.has_vector_multiply:
            raise NotImplementedError(
                "Vector multiplication not defined on this space."
            )
        return self._vector_multiply(x1, x2)

    def to_components(self, x):
        """Maps vectors to components."""
        return self.__to_components(x)

    def from_components(self, c):
        """Maps components to vectors."""
        return self.__from_components(c)

    def basis_vector(self, i):
        """Return the ith basis vector."""
        c = np.zeros(self.dim)
        c[i] = 1
        return self.from_components(c)

    def random(self):
        """Returns a random vector with components drwn from a standard Gaussian distribution."""
        return self.from_components(np.random.randn(self.dim))

    def sample_expectation(self, vectors):
        """
        Given a list of elements in the space, forms their sample variance.
        """
        n = len(vectors)
        all([self.is_element(x) for x in vectors])
        xbar = self.zero
        for x in vectors:
            xbar = self.axpy(1 / n, x, xbar)
        return xbar

    def identity_operator(self):
        """Returns identity operator on the space."""
        return LinearOperator(
            self,
            self,
            lambda x: x,
            dual_mapping=lambda yp: yp,
            adjoint_mapping=lambda y: y,
        )

    def zero_operator(self, codomain=None):
        """Returns zero operator into another space."""
        codomain = self if codomain is None else codomain
        return LinearOperator(
            self,
            codomain,
            lambda x: codomain.zero,
            dual_mapping=lambda yp: self.dual.zero,
            adjoint_mapping=lambda y: self.zero,
        )

    def _dual_to_components(self, xp):
        return xp.components

    def _dual_from_components(self, cp):
        return LinearForm(self, components=cp)

    def __add(self, x, y):
        # Default implementation of vector addition.
        return x + y

    def __subtract(self, x, y):
        # Default implementation of vector subtraction.
        return x - y

    def __multiply(self, a, x):
        # Default implementation of scalar multiplication.
        return a * x.copy()

    def __axpy(self, a, x, y):
        # Default implementation of y -> a*x+y.
        y += a * x
        return y

    def __copy(self, x):
        return x.copy()


class EuclideanSpace(HilbertSpace):
    """
    Euclidean space implemented as an instance of HilbertSpace."""

    def __init__(self, dim):
        """
        Args:
            dim (int): Dimension of the space.
        """

        super().__init__(
            dim,
            lambda x: x,
            lambda x: x,
            self.__inner_product,
            self.__to_dual,
            self.__from_dual,
        )

    def __inner_product(self, x1, x2):
        return np.dot(x1, x2)

    def __to_dual(self, x):
        return self.dual.from_components(x)

    def __from_dual(self, xp):
        cp = self.dual.to_components(xp)
        return self.from_components(cp)

    def __eq__(self, other):
        """
        Overload of equality operator for Euclidean spaces.
        """
        return isinstance(other, EuclideanSpace) and self.dim == other.dim


class Operator:
    """
    Class for operators between two Hilbert spaces.
    """

    def __init__(self, domain, codomain, mapping):
        """
        Args:
            domain (HilbertSpace): Domain of the operator.
            codomain (HilbertSpace): Codomain of the operator.
            mapping (callable): Mapping from domain to codomain.
        """
        self._domain = domain
        self._codomain = codomain
        self.__mapping = mapping

    @property
    def domain(self):
        """Domain of the operator."""
        return self._domain

    @property
    def codomain(self):
        """Codomain of the operator."""
        return self._codomain

    @property
    def is_automorphism(self):
        """True is operator maps a space into itself."""
        return self.domain == self.codomain

    @property
    def is_square(self):
        """True is operator maps a space into itself."""
        return self.domain.dim == self.codomain.dim

    @property
    def linear(self):
        """
        True is the operator is linear.
        """
        return False

    def __call__(self, x):
        """Action of the operator on a vector."""
        return self.__mapping(x)


class LinearOperator(Operator):
    """
    Class for linear operators between two Hilbert spaces.
    """

    def __init__(
        self,
        domain,
        codomain,
        mapping,
        /,
        *,
        adjoint_mapping=None,
        dual_mapping=None,
        thread_safe=False,
        dual_base=None,
        adjoint_base=None,
    ):
        """
        Args:
            domain (HilbertSpace): The domain of the operator.
            codomain (HilbertSpace): The codomain of the operator.
            mapping (callable): Mapping from the domain to codomain.
            dual_mapping (callable): Optional implementation of
                dual operator's action.
            adjoint_mapping (callable): Optional implementation
                of the adjoint operator's action.

            thread_safe (bool): True if the operators action can be
                safely called in parallel. Default is false.
            dual_base (LinearOperator) : Used internally when defining
                dual operators. Should not be set manually.
            adjoint_base (LinearOperator): Used internally when defining
                adjoint operators. Should not be set manually.
        """
        super().__init__(domain, codomain, mapping)
        self._dual_base = dual_base
        self._adjoint_base = adjoint_base
        self._thread_safe = thread_safe

        if dual_mapping is None:
            if adjoint_mapping is None:
                self.__dual_mapping = self._dual_mapping_default
                self.__adjoint_mapping = self._adjoint_mapping_from_dual
            else:
                self.__adjoint_mapping = adjoint_mapping
                self.__dual_mapping = self._dual_mapping_from_adjoint
        else:
            self.__dual_mapping = dual_mapping
            if adjoint_mapping is None:
                self.__adjoint_mapping = self._adjoint_mapping_from_dual
            else:
                self.__adjoint_mapping = adjoint_mapping

    @staticmethod
    def self_dual(domain, mapping):
        """Returns a self-dual operator in terms of its domain and mapping."""
        return LinearOperator(domain, domain.dual, mapping, dual_mapping=mapping)

    @staticmethod
    def self_adjoint(domain, mapping):
        """Returns a self-adjoint operator in terms of its domain and mapping."""
        return LinearOperator(domain, domain, mapping, adjoint_mapping=mapping)

    @staticmethod
    def from_formal_adjoint(domain, codomain, mapping, formal_adjoint):
        """
        Forms a LinearOperator mapping the domain to the codomain. The
        action of the mapping is provided along with that of its formal
        adjoint. The formal adjoint is a mapping from the codomain to the
        domain and is the adjoint of the operator relative to an L2 inner product.
        """

        def dual_mapping(yp):
            cyp = codomain.dual.to_components(yp)
            y = codomain.from_components(cyp)
            x = formal_adjoint(y)
            cx = domain.to_components(x)
            return domain.dual.from_components(cx)

        return LinearOperator(domain, codomain, mapping, dual_mapping=dual_mapping)

    @staticmethod
    def formally_self_adjoint(domain, mapping):
        """
        Forms a LinearOperator on the domain given the mapping
        on the assumption that the operator is formally self-adjoint.
        """
        return LinearOperator.from_formal_adjoint(domain, domain, mapping, mapping)

    @staticmethod
    def from_linear_forms(forms):
        """
        Returns a linear operator into Euclidiean space defined by the tensor
        product of a set of forms with the standard Euclidean basis vectors.

        Args:
            forms ([LinearForms]): A list of linear forms defined on a common domain.

        Returns:
            LinearOperator: The linear operator.

        Notes: The matrix components of the forms are used to define the
               matrix representation of the operator and this is stored internally.
        """
        domain = forms[0].domain
        codomain = EuclideanSpace(len(forms))
        if not all(form.domain == domain for form in forms):
            raise ValueError("Forms need to be defined on a common domain")

        matrix = np.zeros((codomain.dim, domain.dim))
        for i, form in enumerate(forms):
            matrix[i, :] = form.components

        def mapping(x):
            cx = domain.to_components(x)
            cy = matrix @ cx
            return cy

        def dual_mapping(yp):
            cyp = codomain.dual.to_components(yp)
            cxp = matrix.T @ cyp
            return domain.dual.from_components(cxp)

        return LinearOperator(domain, codomain, mapping, dual_mapping=dual_mapping)

    @staticmethod
    def from_matrix(domain, codomain, matrix, /, *, galerkin=False):
        """
        Returns a linear operator defined by its matrix representation.
        By default the standard representation is assumed but the
        Galerkin representation can optinally be used so long as the
        domain and codomain are Hilbert spaces.

        Args:
            domain (HilbertSpace): The domain of the operator.
            codomain (HilbertSpace): The codomain of the operator.
            matrix (matrix-like): The matrix representation of the operator.
            galerkin (bool): True if the Galkerin represention is used.

        Returns:
            LinearOperator: The linear operator.
        """
        assert matrix.shape == (codomain.dim, domain.dim)

        if galerkin:

            def mapping(x):
                cx = domain.to_components(x)
                cyp = matrix @ cx
                yp = codomain.dual.from_components(cyp)
                return codomain.from_dual(yp)

            def adjoint_mapping(y):
                cy = codomain.to_components(y)
                cxp = matrix.T @ cy
                xp = domain.dual.from_components(cxp)
                return domain.from_dual(xp)

            return LinearOperator(
                domain,
                codomain,
                mapping,
                adjoint_mapping=adjoint_mapping,
            )

        else:

            def mapping(x):
                cx = domain.to_components(x)
                cy = matrix @ cx
                return codomain.from_components(cy)

            def dual_mapping(yp):
                cyp = codomain.dual.to_components(yp)
                cxp = matrix.T @ cyp
                return domain.dual.from_components(cxp)

            return LinearOperator(domain, codomain, mapping, dual_mapping=dual_mapping)

    @staticmethod
    def self_adjoint_from_matrix(domain, matrix):
        """
        Forms a self-adjoint operator from its Galerkin matrix representation.
        """

        def mapping(x):
            cx = domain.to_components(x)
            cyp = matrix @ cx
            yp = domain.dual.from_components(cyp)
            return domain.from_dual(yp)

        return LinearOperator.self_adjoint(domain, mapping)

    @staticmethod
    def from_tensor_product(domain, codomain, vector_pairs, /, *, weights=None):
        """
        Forms a LinearOperator between Hilbert spaces from the tensor product
        of a list of pairs of vectors.

        Args:
            domain (HilbertSpace): Domain for the linear operator.
            codomain (HilbertSpace): Codomain for the linear operator.
            vector_pairs ([[codomain vector, domain vector]]): A list of pairs of vectors
                from which the tensor product is to be constructed.
            weights ([float]): Optional list of weights for the terms in the tensor
               product. If none is provided default weights of one are used.
        """

        assert all(domain.is_element(vector) for _, vector in vector_pairs)
        assert all(codomain.is_element(vector) for vector, _ in vector_pairs)

        if weights is None:
            _weights = [1 for _ in vector_pairs]
        else:
            _weights = weights

        def mapping(x):
            y = codomain.zero
            for left, right, weight in zip(vector_pairs, _weights):
                product = domain.inner_product(right, x)
                y = codomain.axpy(weight * product, left, y)
            return y

        def adjoint_mapping(y):
            x = domain.zero
            for left, right, weight in zip(vector_pairs, _weights):
                product = codomain.inner_product(left, y)
                x = domain.axpy(weight * product, right, x)
            return x

        return LinearOperator(
            domain, codomain, mapping, adjoint_mapping=adjoint_mapping
        )

    @staticmethod
    def self_adjoint_from_tensor_product(domain, vectors, /, *, weights=None):
        """
        Forms a self-adjoint LinearOperator on a Hilbert space from
        the tensor product of a list of vectors.

        Args:
            domain (HilbertSpace): Domain for the linear operator.
            vectors ([domain vector]): A list of vectors from which
                the tensor product is to be constructed.
            weights [float]: Optional list of weights for the terms in the tensor
                product. If none is provided default weights of one are used.
        """

        assert all(domain.is_element(vector) for vector in vectors)

        if weights is None:
            _weights = [1 for _ in vectors]
        else:
            _weights = weights

        def mapping(x):
            y = domain.zero
            for vector, weight in zip(vectors, _weights):
                product = domain.inner_product(vector, x)
                y = domain.axpy(weight * product, vector, y)
            return y

        return LinearOperator.self_adjoint(domain, mapping)

    @property
    def linear(self):
        # Overide of method from base class.
        return True

    @property
    def dual(self):
        """The dual of the operator."""
        if self._dual_base is None:
            return LinearOperator(
                self.codomain.dual,
                self.domain.dual,
                self.__dual_mapping,
                dual_mapping=self,
                dual_base=self,
            )
        else:
            return self._dual_base

    @property
    def adjoint(self):
        """The adjoint of the operator."""
        if self._adjoint_base is None:
            return LinearOperator(
                self.codomain,
                self.domain,
                self.__adjoint_mapping,
                adjoint_mapping=self,
                adjoint_base=self,
            )
        else:
            return self._adjoint_base

    @property
    def thread_safe(self):
        """
        Returns True if operator is thread safe.
        """
        return self._thread_safe

    def matrix(self, /, *, dense=False, galerkin=False):
        """Return matrix representation of the operator."""
        if dense:
            return self._compute_dense_matrix(galerkin)
        else:

            # Implement matrix-vector and transposed-matrix-vector products
            if galerkin:

                def matvec(cx):
                    x = self.domain.from_components(cx)
                    y = self(x)
                    yp = self.codomain.to_dual(y)
                    return self.codomain.dual.to_components(yp)

                def rmatvec(cy):
                    y = self.codomain.from_components(cy)
                    x = self.adjoint(y)
                    xp = self.domain.to_dual(x)
                    return self.domain.dual.to_components(xp)

            else:

                def matvec(cx):
                    x = self.domain.from_components(cx)
                    y = self(x)
                    return self.codomain.to_components(y)

                def rmatvec(cyp):
                    yp = self.codomain.dual.from_components(cyp)
                    xp = self.dual(yp)
                    return self.domain.dual.to_components(xp)

            # Implement matrix-matrix and transposed-matrix-matrix products
            def matmat(xmat):
                n, k = xmat.shape
                assert n == self.domain.dim
                ymat = np.zeros((self.codomain.dim, k))
                for j in range(k):
                    cx = xmat[:, j]
                    ymat[:, j] = matvec(cx)
                return ymat

            def rmatmat(ymat):
                m, k = ymat.shape
                assert m == self.codomain.dim
                xmat = np.zeros((self.domain.dim, k))
                for j in range(k):
                    cy = ymat[:, j]
                    xmat[:, j] = rmatvec(cy)
                return xmat

            # Return the scipy LinearOperator
            return ScipyLinOp(
                (self.codomain.dim, self.domain.dim),
                matvec=matvec,
                rmatvec=rmatvec,
                matmat=matmat,
                rmatmat=rmatmat,
            )

    def random_svd(
        self, rank, /, *, power=0, galerkin=False, rtol=1e-3, method="fixed"
    ):
        """
        Returns an approximate singular value decomposition of an operator using
        the randomised method of Halko et al. 2011.

        Let X and Y be the domain and codomain of the operator, A, and E be Euclidean
        space of dimension rank. The factorisation then takes the form

        A = L D R

        where the factors map:

        R : X --> E
        D : E --> E
        L : E --> Y

        with D diagonal and comprises the singular values for the operator.

        Args:

            rank (int) : rank for the decomposition.
            power (int) : exponent within the power iterations.
            galerkin (bool) : If true, use the Galerkin representation of the
                operator. Only possible if the operator maps between Hilbert spaces.

        Returns:
            LinearOperator: The left factor, L.
            DiagonalLinearOperator: The diagonal factor, D.
            LinearOperator: The right factor, R.

        """
        matrix = self.matrix(galerkin=galerkin)
        m, n = matrix.shape
        k = min(m, n)
        rank = rank if rank <= k else k

        if method == "fixed":
            qr_factor = fixed_rank_random_range(matrix, rank, power=power)
        elif method == "variable":
            qr_factor = variable_rank_random_range(matrix, rank, power=power, rtol=rtol)
        else:
            raise ValueError("Invalid method selected")

        left_factor, singular_values, right_factor_transposed = random_svd(
            matrix, qr_factor
        )

        euclidean = EuclideanSpace(qr_factor.shape[1])
        diagonal = DiagonalLinearOperator(euclidean, euclidean, singular_values)

        if galerkin:

            def right_mapping(x):
                cx = self.domain.to_components(x)
                return right_factor_transposed @ cx

            def right_mapping_adjoint(cx):
                cxp = right_factor_transposed.T @ cx
                xp = self.domain.dual.from_components(cxp)
                return self.domain.from_dual(xp)

            right = LinearOperator(
                self.domain,
                euclidean,
                right_mapping,
                adjoint_mapping=right_mapping_adjoint,
            )

            def left_mapping(cx):
                cyp = left_factor @ cx
                yp = self.codomain.dual.from_components(cyp)
                return self.codomain.from_dual(yp)

            def left_mapping_adjoint(y):
                cy = self.codomain.to_components(y)
                return left_factor.T @ cy

            left = LinearOperator(
                euclidean,
                self.codomain,
                left_mapping,
                adjoint_mapping=left_mapping_adjoint,
            )

        else:

            def right_mapping(x):
                cx = self.domain.to_components(x)
                return right_factor_transposed @ cx

            def right_mapping_dual(cp):
                c = euclidean.from_dual(cp)
                cxp = right_factor_transposed.T @ c
                return self.domain.dual.from_components(cxp)

            right = LinearOperator(
                self.domain, euclidean, right_mapping, dual_mapping=right_mapping_dual
            )

            def left_mapping(c):
                cy = left_factor @ c
                return self.codomain.from_components(cy)

            def left_mapping_dual(yp):
                cpy = self.codomain.dual.to_components(yp)
                c = left_factor.T @ cpy
                return euclidean.to_dual(c)

            left = LinearOperator(
                euclidean, self.codomain, left_mapping, dual_mapping=left_mapping_dual
            )

        # Return the factors.
        return left, diagonal, right

    def random_eig(self, rank, /, *, power=0, rtol=1e-3, method="fixed"):
        """
        Returns an approximate eigenvalue decomposition of a self-adjoint operator using
        the randomised method of Halko et al. 2011.

        Let X  the domain the operator, A, and E be Euclidean space of dimension rank.
        The factorisation then takes the form

        A = U D U*

        where the factors map:

        U : E --> X
        D : E --> E

        with D diagonal and comprises the eigenvalues of the operator.

        If the diagonal values are non-zero, we can also factor an
        approximation to the inverse mapping as:

        A^{-1} = V D^{-1} V*

        where V = I I* U with I the coordinate_inclusion mapping on the Hilbert space.


        Args:
            rank (int) : rank for the decomposition.
            power (int) : exponent within the power iterations.
            inverse (bool): If true, return the decomposition for
                the inverse operator.

        Returns:
            LinearOperator: The factor, U.
            DiagonalLinearOperator: The diagonal factor, D.
        """

        assert self.is_automorphism
        matrix = self.matrix(galerkin=True)
        m, n = matrix.shape
        k = min(m, n)
        rank = rank if rank <= k else k

        if method == "fixed":
            qr_factor = fixed_rank_random_range(matrix, rank, power=power)
        elif method == "variable":
            qr_factor = variable_rank_random_range(matrix, rank, power=power, rtol=rtol)
        else:
            raise ValueError("Invalid method selected")

        eigenvectors, eigenvalues = random_eig(matrix, qr_factor)
        euclidean = EuclideanSpace(qr_factor.shape[1])
        diagonal = DiagonalLinearOperator(euclidean, euclidean, eigenvalues)

        def mapping(c):
            cyp = eigenvectors @ c
            yp = self.domain.dual.from_components(cyp)
            return self.domain.from_dual(yp)

        def adjoint_mapping(x):
            cx = self.domain.to_components(x)
            return eigenvectors.T @ cx

        expansion = LinearOperator(
            euclidean, self.domain, mapping, adjoint_mapping=adjoint_mapping
        )

        return expansion, diagonal

    def random_cholesky(self, rank, /, *, power=0, rtol=1e-3, method="fixed"):
        """
        Returns an approximate Cholesky decomposition of a positive-definite and
        self-adjoint operator using the randomised method of Halko et al. 2011.

        Let X  the domain the operator, A, and E be Euclidean space of dimension rank.
        The factorisation then takes the form

        A = F F*

        where F : E --> X.

        Args:

            rank (int) : rank for the decomposition.
            power (int) : exponent within the power iterations.

        Returns:
            LinearOperator : The Cholesky factor, F

        """
        assert self.is_automorphism
        matrix = self.matrix(galerkin=True)
        m, n = matrix.shape
        k = min(m, n)
        rank = rank if rank <= k else k

        if method == "fixed":
            qr_factor = fixed_rank_random_range(matrix, rank, power=power)
        elif method == "variable":
            qr_factor = variable_rank_random_range(matrix, rank, power=power, rtol=rtol)
        else:
            raise ValueError("Invalid method selected")

        cholesky_factor = random_cholesky(matrix, qr_factor)

        def mapping(x):
            cyp = cholesky_factor @ x
            yp = self.codomain.dual.from_components(cyp)
            return self.codomain.from_dual(yp)

        def adjoint_mapping(y):
            x = self.codomain.to_components(y)
            return cholesky_factor.T @ x

        return LinearOperator(
            EuclideanSpace(qr_factor.shape[1]),
            self.domain,
            mapping,
            adjoint_mapping=adjoint_mapping,
        )

    def _dual_mapping_default(self, yp):
        # Default implementation of the dual mapping.
        return LinearForm(self.domain, mapping=lambda x: yp(self(x)))

    def _dual_mapping_from_adjoint(self, yp):
        # Dual mapping in terms of the adjoint.
        y = self.codomain.from_dual(yp)
        x = self.__adjoint_mapping(y)
        return self.domain.to_dual(x)

    def _adjoint_mapping_from_dual(self, y):
        # Adjoing mapping in terms of the dual.
        yp = self.codomain.to_dual(y)
        xp = self.__dual_mapping(yp)
        return self.domain.from_dual(xp)

    def _compute_dense_matrix(self, galerkin=False):
        # Compute the matrix representation in dense form.
        matrix = np.zeros((self.codomain.dim, self.domain.dim))
        a = self.matrix(galerkin=galerkin)
        cx = np.zeros(self.domain.dim)
        for i in range(self.domain.dim):
            cx[i] = 1
            matrix[:, i] = (a @ cx)[:]
            cx[i] = 0
        return matrix

    def __neg__(self):
        """negative unary"""
        domain = self.domain
        codomain = self.codomain

        def mapping(x):
            return codomain.negative(self(x))

        def adjoint_mapping(y):
            return domain.negative(self.adjoint(y))

        return LinearOperator(
            domain,
            codomain,
            mapping,
            adjoint_mapping=adjoint_mapping,
        )

    def __mul__(self, a):
        """Multiply by a scalar."""
        domain = self.domain
        codomain = self.codomain

        def mapping(x):
            return codomain.multiply(a, self(x))

        def adjoint_mapping(y):
            return domain.multiply(a, self.adjoint(y))

        return LinearOperator(
            domain,
            codomain,
            mapping,
            adjoint_mapping=adjoint_mapping,
        )

    def __rmul__(self, a):
        """Multiply by a scalar."""
        return self * a

    def __truediv__(self, a):
        """Divide by scalar."""
        return self * (1 / a)

    def __add__(self, other):
        """Add another operator."""
        domain = self.domain
        codomain = self.codomain

        def mapping(x):
            return codomain.add(self(x), other(x))

        def adjoint_mapping(y):
            return domain.add(self.adjoint(y), other.adjoint(y))

        return LinearOperator(
            domain,
            codomain,
            mapping,
            adjoint_mapping=adjoint_mapping,
        )

    def __sub__(self, other):
        """Subtract another operator."""
        domain = self.domain
        codomain = self.codomain

        def mapping(x):
            return codomain.subtract(self(x), other(x))

        def adjoint_mapping(y):
            return domain.subtract(self.adjoint(y), other.adjoint(y))

        return LinearOperator(
            domain,
            codomain,
            mapping,
            adjoint_mapping=adjoint_mapping,
        )

    def __matmul__(self, other):
        """Compose with another operator."""
        domain = other.domain
        codomain = self.codomain

        def mapping(x):
            return self(other(x))

        def adjoint_mapping(y):
            return other.adjoint(self.adjoint(y))

        return LinearOperator(
            domain,
            codomain,
            mapping,
            adjoint_mapping=adjoint_mapping,
        )

    def __str__(self):
        """Print the operator as its dense matrix representation."""
        return self.matrix(dense=True).__str__()


class DiagonalLinearOperator(LinearOperator):
    """
    Class for Linear operators whose matrix representation is diagonal.
    """

    def __init__(self, domain, codomain, diagonal_values, /, *, galerkin=False):
        """
        Args:
            domain (HilbertSpace): The domain of the operator.
            codoomain (HilbertSpace): The codomain of the operator.
            diagonal_values (numpy vector): Diagonal values for the
                operator's matrix representation.
            galerkin (bool): true is galerkin representation is used.
        """

        assert domain.dim == codomain.dim
        assert domain.dim == len(diagonal_values)
        self._diagonal_values = diagonal_values
        matrix = diags([diagonal_values], [0])
        operator = LinearOperator.from_matrix(
            domain, codomain, matrix, galerkin=galerkin
        )
        super().__init__(
            operator.domain,
            operator.codomain,
            operator,
            adjoint_mapping=operator.adjoint,
        )

    @property
    def diagonal_values(self):
        """
        Return the diagonal values.
        """
        return self._diagonal_values

    def function(self, f):
        """
        Returns a function, f, of the operator. It is assumed that the
        function is well-defined on the diagonal values.
        """
        diagonal_values = np.array([f(x) for x in self.diagonal_values])
        return DiagonalLinearOperator(self.domain, self.codomain, diagonal_values)

    @property
    def inverse(self):
        """
        return the inverse operator. Valid only if diagonal values
        are non-zero.
        """
        assert all(val != 0 for val in self.diagonal_values)
        return self.function(lambda x: 1 / x)

    @property
    def sqrt(self):
        """
        Returns the square root of the operator. Valid only if diagonal values
        are non-negative.
        """
        assert all(val >= 0 for val in self._diagonal_values)
        return self.function(np.sqrt)


class LinearForm:
    """
    Class for linear forms on a Hilbert space. Can be specified by its
    action or through its components.
    """

    def __init__(self, domain, /, *, mapping=None, components=None):
        """
        Args:
            domain (HilbertSpace): Domain of the linear form.
            mapping (callable | None): A functor that performs the action
                of the linear form on a vector.
            matrix (MatrixLike | None): The matrix representation of the
                form, this having shape (1,dim) with dim the dimension of
                the domain.
        """

        self._domain = domain
        if components is None:
            self._components = None
            if mapping is None:
                raise AssertionError("Neither mapping nor components specified.")
            else:
                self._mapping = mapping
        else:
            self._components = components
            if mapping is None:
                self._mapping = self._mapping_from_components
            else:
                self._mapping = mapping

    @staticmethod
    def from_linear_operator(operator):
        """
        Form a linear form from an linear operator mapping onto one-dimensional Euclidean space.
        """
        assert operator.codomain == EuclideanSpace(1)
        return LinearForm(operator.domain, mapping=lambda x: operator(x)[0])

    @property
    def domain(self):
        """Return the form the domain is defined on"""
        return self._domain

    @property
    def components_stored(self):
        """True is the form has its components stored."""
        return self._components is not None

    @property
    def components(self):
        """Return the components of the form."""
        if self.components_stored:
            return self._components
        else:
            self.store_components()
            return self.components

    def store_components(self):
        """Compute and store the forms components."""
        if not self.components_stored:
            self._components = np.zeros(self.domain.dim)
            cx = np.zeros(self.domain.dim)
            for i in range(self.domain.dim):
                cx[i] = 1
                x = self.domain.from_components(cx)
                self._components[i] = self(x)
                cx[i] = 0

    @property
    def as_linear_operator(self):
        """
        Return the linear form as a LinearOperator.
        """
        return LinearOperator(
            self.domain,
            EuclideanSpace(1),
            lambda x: np.array([self(x)]),
            dual_mapping=lambda y: y * self,
        )

    def __call__(self, x):
        """Action of the form on a vector"""
        return self._mapping(x)

    def __neg__(self):
        """negative unary"""
        if self.components_stored:
            return LinearForm(self.domain, components=-self._components)
        else:
            return LinearForm(self.domain, mapping=lambda x: -self(x))

    def __mul__(self, a):
        """Multiply by a scalar."""
        if self.components_stored:
            return LinearForm(self.domain, components=a * self._components)
        else:
            return LinearForm(self.domain, mapping=lambda x: a * self(x))

    def __rmul__(self, a):
        """Multiply by a scalar."""
        return self * a

    def __truediv__(self, a):
        """Divide by scalar."""
        return self * (1 / a)

    def __add__(self, other):
        """Add another form."""
        if self.components_stored and other.components_stored:
            return LinearForm(
                self.domain, components=self.components + other.components
            )
        else:
            return LinearForm(self.domain, mapping=lambda x: self(x) + other(x))

    def __sub__(self, other):
        """Subtract another form."""
        if self.components_stored and other.components_stored:
            return LinearForm(
                self.domain, components=self.components - other.components
            )
        else:
            return LinearForm(self.domain, mapping=lambda x: self(x) - other(x))

    def __str__(self):
        return self.components.__str__()

    def _mapping_from_components(self, x):
        # Implement the action of the form using its components.
        return np.dot(self._components, self.domain.to_components(x))



"""
Module for random matrix factorisations. 
"""

import numpy as np
from scipy.linalg import (
    cho_factor,
    solve_triangular,
    eigh,
    svd,
    qr,
)


def fixed_rank_random_range(matrix, rank, power=0):
    """
    Forms the fixed-rank approximation to the range of a matrix using
    a random-matrix method.

    Args:
        matrix (matrix-like): (m,n)-matrix whose range is to be approximated.
        rank (int): The desired rank. Must be greater than 1.
        power (int): The exponent to use within the power iterations.

    Returns:
        matrix: A (m,rank)-matrix whose columns are orthonormal and
            whose span approximates the desired range.

    Notes:
        The input matrix can be a numpy array or a scipy LinearOperator. In the latter case,
        it requires the the matmat, and rmatmat methods have been implemented.

        This method is based on Algorithm 4.4 in Halko et. al. 2011
    """

    m, n = matrix.shape
    random_matrix = np.random.randn(n, rank)

    product_matrix = matrix @ random_matrix
    qr_factor, _ = qr(product_matrix, overwrite_a=True, mode="economic")

    for _ in range(power):
        tilde_product_matrix = matrix.T @ qr_factor
        tilde_qr_factor, _ = qr(tilde_product_matrix, overwrite_a=True, mode="economic")
        product_matrix = matrix @ tilde_qr_factor
        qr_factor, _ = qr(product_matrix, overwrite_a=True, mode="economic")

    return qr_factor


def variable_rank_random_range(matrix, rank, /, *, power=0, rtol=1e-6):
    """
    Forms the variable-rank approximation to the range of a matrix using
    a random-matrix method.

    Args:
        matrix (matrix-like): (m,n)-matrix whose range is to be approximated.

    Returns:
        matrix: A (m,rank)-matrix whose columns are orthonormal and
            whose span approximates the desired range.

    Notes:
        The input matrix can be a numpy array or a scipy LinearOperator. In the latter case,
        it requires the the matmat, and rmatmat methods have been implemented.

        This method is based on Algorithm 4.5 in Halko et. al. 2011
    """

    m, n = matrix.shape

    random_vectors = [np.random.randn(n) for _ in range(rank)]
    ys = [matrix @ x for x in random_vectors]
    basis_vectors = []

    def projection(xs, y):
        ps = [np.dot(x, y) for x in xs]
        for p, x in zip(ps, xs):
            y -= p * x
        return y

    norm = max(np.linalg.norm(y) for y in ys)

    tol = rtol * norm / (10 * np.sqrt(2 / np.pi))
    error = 2 * tol
    j = -1
    while error > tol:
        j += 1

        ys[j] = projection(basis_vectors, ys[j])
        ys[j] /= np.linalg.norm(ys[j])
        basis_vectors.append(ys[j])

        y = matrix @ np.random.randn(n)
        y = projection(basis_vectors, y)
        ys.append(y)

        for i in range(j + 1, j + rank):
            p = np.dot(basis_vectors[j], ys[i])
            ys[i] -= p * basis_vectors[j]

        error = max(np.linalg.norm(ys[i]) for i in range(j + 1, j + rank + 1))

        if j > min(n, m):
            raise RuntimeError("Convergence has failed")

    qr_factor = np.column_stack(basis_vectors)

    return qr_factor


def random_svd(matrix, qr_factor):
    """
    Given a matrix, A,  and a low-rank approximation to its range, Q,
    this function returns the approximate SVD factors, (U, S, Vh)
    such that A ~ U @ S @ VT where S is diagonal.

    Based on Algorithm 5.1 of Halko et al. 2011
    """
    small_matrix = qr_factor.T @ matrix
    left_factor, diagonal_factor, right_factor_transposed = svd(
        small_matrix, full_matrices=False, overwrite_a=True
    )
    return (
        qr_factor @ left_factor,
        diagonal_factor,
        right_factor_transposed,
    )


def random_eig(matrix, qr_factor):
    """
    Given a symmetric matrix, A,  and a low-rank approximation to its range, Q,
    this function returns the approximate eigen-decomposition, (U, S)
    such that A ~ U @ S @ U.T where S is diagonal.

    Based on Algorithm 5.3 of Halko et al. 2011
    """
    m, n = matrix.shape
    assert m == n
    small_matrix = qr_factor.T @ matrix @ qr_factor
    eigenvalues, eigenvectors = eigh(small_matrix, overwrite_a=True)
    return qr_factor @ eigenvectors, eigenvalues


def random_cholesky(matrix, qr_factor):
    """
    Given a symmetric and positive-definite matrix, A,  along with a low-rank
    approximation to its range, Q, this function returns the approximate
    Cholesky factorisation A ~ F F*.

    Based on Algorithm 5.5 of Halko et al. 2011
    """
    small_matrix_1 = matrix @ qr_factor
    small_matrix_2 = qr_factor.T @ small_matrix_1
    factor, lower = cho_factor(small_matrix_2, overwrite_a=True)
    identity_operator = np.identity(factor.shape[0])
    inverse_factor = solve_triangular(
        factor, identity_operator, overwrite_b=True, lower=lower
    )
    return small_matrix_1 @ inverse_factor



"""
Module for the Bayesian approaches to Bayesian inverse problems.
"""

import numpy as np
from scipy.linalg import cho_factor, solve_triangular
from pygeoinf.hilbert_space import LinearOperator, EuclideanSpace
from pygeoinf.linear_solvers import IterativeLinearSolver
from pygeoinf.gaussian_measure import GaussianMeasure
from pygeoinf.inversion import Inversion


class LinearBayesianInversion(Inversion):
    """
    Class for solving a linear inverse problem Bayesian methods assuming Gaussian priors and errors.
    """

    def __init__(self, forward_problem, model_prior_measure):
        """
        Args:
            forward_problem (LinearForwardProblem): The forward problem.
            model_prior_measure (GaussianMeasure): The prior measure on the data.
        """
        super().__init__(forward_problem)
        self._model_prior_measure = model_prior_measure

    @property
    def forward_problem(self):
        """
        Return the forward problem.
        """
        return self._forward_problem

    @property
    def model_prior_measure(self):
        """Return the model prior measure."""
        return self._model_prior_measure

    @property
    def normal_operator(self):
        """
        Returns the data-space normal operator.
        """
        forward_operator = self.forward_problem.forward_operator
        prior_model_covariance = self.model_prior_measure.covariance

        if self.forward_problem.data_error_measure_set:
            return (
                forward_operator @ prior_model_covariance @ forward_operator.adjoint
                + self.forward_problem.data_error_measure.covariance
            )
        else:
            return forward_operator @ prior_model_covariance @ forward_operator.adjoint

    def data_prior_measure(self):
        """
        Return the prior distribution on the data
        """
        if self.forward_problem.data_error_measure_set:
            return (
                self.model_prior_measure.affine_mapping(
                    operator=self.forward_problem.forward_operator
                )
                + self.forward_problem.data_error_measure
            )
        else:
            return self.model_prior_measure.affine_mapping(
                operator=self.forward_problem.forward_operator
            )

    def model_posterior_measure(self, data, solver, /, *, preconditioner=None):
        """
        Returns the posterior measure on the model space given the data.

        Args:
            data (data-space vector): The observed data.
            solver (LinearSolver): A linear solver for the normal equations.
            preconditioner (LinearSolver): A preconditioner for use in solving
                the normal equations.

        Returns:
            GaussianMeasure: The posterior measure.

        Notes:
            The posterior measure does not have a sampling method set. If required,
            this should be set directly afterwards.
        """

        data_space = self.data_space
        model_space = self.model_space
        forward_operator = self.forward_problem.forward_operator
        prior_model_covariance = self.model_prior_measure.covariance
        normal_operator = self.normal_operator

        if isinstance(solver, IterativeLinearSolver):
            inverse_normal_operator = solver(
                normal_operator, preconditioner=preconditioner
            )
        else:
            inverse_normal_operator = solver(normal_operator)

        shifted_data = data_space.subtract(
            data, forward_operator(self.model_prior_measure.expectation)
        )
        shifted_data = data_space.subtract(
            shifted_data, self.forward_problem.data_error_measure.expectation
        )

        expectation = (
            prior_model_covariance @ forward_operator.adjoint @ inverse_normal_operator
        )(shifted_data)
        expectation = model_space.add(expectation, self.model_prior_measure.expectation)

        covariance = (
            prior_model_covariance
            - prior_model_covariance
            @ forward_operator.adjoint
            @ inverse_normal_operator
            @ forward_operator
            @ prior_model_covariance
        )

        return GaussianMeasure(covariance=covariance, expectation=expectation)


class LinearBayesianInference(LinearBayesianInversion):
    """Class for solving Bayesian inference problems."""

    def __init__(self, forward_problem, model_prior_measure, property_operator):
        """
        Args:
            forward_problem (LinearForwardProblem): The forward problem.
            model_prior_measure (GaussianMeasure): The prior measure on the data.
            property_operator (LinearOperator): The property operator.

        """
        super().__init__(forward_problem, model_prior_measure)
        assert property_operator.domain == self.forward_problem.model_space
        self._property_operator = property_operator

    @property
    def property_space(self):
        """Return the property space."""
        return self._property_operator.codomain

    @property
    def property_operator(self):
        """Return the property operator."""
        return self._property_operator

    def property_prior_measure(self):
        """Return the prior measure on the property space."""
        return self.model_prior_measure.affine_mapping(operator=self.property_operator)

    def property_posterior_measure(self, data, solver, /, *, preconditioner=None):
        """
        Returns the posterior measure on the property space given the data.

        Args:
            data (data-space vector): The observed data.
            solver (LinearSolver): A linear solver for the normal equations.
            preconditioner (LinearSolver): A preconditioner for use in solving
                the normal equations.

        Returns:
            GaussianMeasure: The posterior measure.

        Notes:
            The posterior measure does not have a sampling method set.
        """
        pi = self.model_posterior_measure(
            data,
            solver,
            preconditioner=preconditioner,
        )
        return pi.affine_mapping(operator=self.property_operator)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm


def plot_colourline(x, y, c):
    fig = plt.figure(1, figsize=(5, 5))
    col = cm.jet((c - np.min(c)) / (np.max(c) - np.min(c)))
    ax = plt.gca()
    for i in np.arange(len(x) - 1):
        ax.plot([x[i], x[i + 1]], [y[i], y[i + 1]], c=col[i])
        im = ax.scatter(x, y, c=c, s=0, cmap=cm.jet)
    fig.colorbar(im)
    plt.show()



"""
Module for direct sums of Hilbert spaces and related constructions. 
"""

from abc import ABC, abstractmethod
import numpy as np
from pygeoinf.hilbert_space import HilbertSpace, LinearOperator, LinearForm


class HilbertSpaceDirectSum(HilbertSpace):
    """
    Class for Hilbert spaces formed from a direct sum of a list of Hilbert spaces.

    Instances are formed by providing a list of HilbertSpaces. Along with the usual
    HilbertSpace methods, this class implements projection and inclusion operators
    onto the subspaces. And also the canonical injection from the direct sum of
    the dual spaces into the dual of the direct sum.
    """

    def __init__(self, spaces):
        """
        Args:
            spaces ([HilbertSpaces]): A list of Hilbert spaces whos direct sum is to be formed.
        """
        self._spaces = spaces
        dim = sum([space.dim for space in spaces])
        super().__init__(
            dim,
            self.__to_components,
            self.__from_components,
            self.__inner_product,
            self.__to_dual,
            self.__from_dual,
            add=self.__add,
            subtract=self.__subtract,
            multiply=self.__multiply,
            axpy=self.__axpy,
            copy=self.__copy,
        )

    #######################################################
    #                    Public methods                   #
    #######################################################

    @property
    def subspaces(self):
        """
        Return a list of the subspaces.
        """
        return self._spaces

    @property
    def number_of_subspaces(self):
        """
        Returns the number of subspaces in the direct sum.
        """
        return len(self.subspaces)

    def subspace(self, i):
        """
        Returns the ith subspace.
        """
        return self.subspaces[i]

    def subspace_projection(self, i):
        """
        Returns the projection operator onto the ith subspace.
        """
        return LinearOperator(
            self,
            self.subspaces[i],
            lambda xs: self._subspace_projection_mapping(i, xs),
            adjoint_mapping=lambda x: self._subspace_inclusion_mapping(i, x),
        )

    def subspace_inclusion(self, i):
        """
        Returns the inclusion operator from the ith space.
        """
        return LinearOperator(
            self.subspaces[i],
            self,
            lambda x: self._subspace_inclusion_mapping(i, x),
            adjoint_mapping=lambda xs: self._subspace_projection_mapping(i, xs),
        )

    def canonical_dual_isomorphism(self, xps):
        """
        Maps a direct sum of dual-subspace vectors to the dual vector.
        """
        if len(xps) != self.number_of_subspaces:
            raise ValueError("Not a valid vector")
        return LinearForm(
            self,
            mapping=lambda x: sum(
                xp(self.subspace_projection(i)(x)) for i, xp in enumerate(xps)
            ),
        )

    def canonical_dual_inverse_isomorphism(self, xp):
        """
        Maps a dual vector to the direct sum of the dual-subspace vectors.
        """
        return [
            LinearForm(space, mapping=lambda x, j=i: xp(self.subspace_inclusion(j)(x)))
            for i, space in enumerate(self.subspaces)
        ]

    #######################################################
    #                   Private methods                   #
    #######################################################

    def __to_components(self, xs):
        # Local implementation of to component mapping.
        cs = [space.to_components(x) for space, x in zip(self._spaces, xs)]
        return np.concatenate(cs, 0)

    def __from_components(self, c):
        # Local implementation of from component mapping.
        xs = []
        i = 0
        for space in self._spaces:
            j = i + space.dim
            x = space.from_components(c[i:j])
            xs.append(x)
            i = j
        return xs

    def __inner_product(self, x1s, x2s):
        # Local implementation of inner product.
        return sum(
            space.inner_product(x1, x2) for space, x1, x2 in zip(self._spaces, x1s, x2s)
        )

    def __to_dual(self, xs):
        # Local implementation of the mapping to the dual space.
        if len(xs) != self.number_of_subspaces:
            raise ValueError("Not a valid vector")
        return self.canonical_dual_isomorphism(
            [space.to_dual(x) for space, x in zip(self._spaces, xs)]
        )

    def __from_dual(self, xp):
        # Local implementation of the mapping from the dual space.
        xps = self.canonical_dual_inverse_isomorphism(xp)
        return [space.from_dual(xip) for space, xip in zip(self._spaces, xps)]

    def __add(self, xs, ys):
        # Local implementation of add.
        return [space.add(x, y) for space, x, y in zip(self._spaces, xs, ys)]

    def __subtract(self, xs, ys):
        # Local implementation of subtract.
        return [space.subtract(x, y) for space, x, y in zip(self._spaces, xs, ys)]

    def __multiply(self, a, xs):
        # Local implementation of multiply.
        return [space.multiply(a, x) for space, x in zip(self._spaces, xs)]

    def __axpy(self, a, xs, ys):
        # Local implementation of axpy.
        return [space.axpy(a, x, y) for space, x, y in zip(self._spaces, xs, ys)]

    def __copy(self, xs):
        return [space.copy(x) for space, x in zip(self._spaces, xs)]

    def _subspace_projection_mapping(self, i, xs):
        # Implementation of the projection mapping onto ith space.
        return xs[i]

    def _subspace_inclusion_mapping(self, i, x):
        # Implementation of the inclusion mapping from ith space.
        return [x if j == i else space.zero for j, space in enumerate(self._spaces)]


class BlockStructure(ABC):
    """
    Base class for block operators.
    """

    def __init__(self, row_dim, col_dim):
        """
        Args:

            row_dim (int): Number of rows in block structure.
            col_dim (int): Number of columns in row structure.
        """
        self._row_dim = row_dim
        self._col_dim = col_dim

    @property
    def row_dim(self):
        """
        Returns the number of rows in block operator.
        """
        return self._row_dim

    @property
    def col_dim(self):
        """
        Returns the number of columns in block operator.
        """
        return self._col_dim

    @abstractmethod
    def block(self, i, j):
        """
        Return the ith block
        """

    def _check_block_indices(self, i, j):
        # Check block indices are in range.
        if not (0 <= i < self.row_dim):
            raise ValueError("Row index out of range.")
        if not (0 <= j < self.col_dim):
            raise ValueError("Column index out of range.")


class BlockLinearOperator(LinearOperator, BlockStructure):
    """
    Class for linear operators acting between direct sums of Hilbert spaces that are defined
    in a blockwise manner.

    Instances are formed from lists of list of LinearOperators which give the sub-blocks in
    a row major ordering.
    """

    def __init__(self, blocks):

        # Check and form the list of domains and codomains.
        domains = [operator.domain for operator in blocks[0]]
        codomains = []
        for row in blocks:
            assert domains == [operator.domain for operator in row]
            codomain = row[0].codomain
            assert all(operator.codomain == codomain for operator in row)
            codomains.append(codomain)

        domain = HilbertSpaceDirectSum(domains)
        codomain = HilbertSpaceDirectSum(codomains)

        self._domains = domains
        self._codomains = codomains
        self._blocks = blocks
        row_dim = len(blocks)
        col_dim = len(blocks[0])

        super().__init__(
            domain,
            codomain,
            self.__mapping,
            adjoint_mapping=self.__adjoint_mapping,
        )
        BlockStructure.__init__(self, row_dim, col_dim)

    def block(self, i, j):
        """
        Returns the operator in the (i,j)th sub-block.
        """
        assert 0 <= i < self.row_dim
        assert 0 <= j < self.col_dim
        return self._blocks[i][j]

    def __mapping(self, xs):
        ys = []
        for i in range(self.row_dim):
            codomain = self._codomains[i]
            y = codomain.zero
            for j in range(self.col_dim):
                a = self.block(i, j)
                y = codomain.axpy(1, a(xs[j]), y)
            ys.append(y)
        return ys

    def __adjoint_mapping(self, ys):
        xs = []
        for j in range(self.col_dim):
            domain = self._domains[j]
            x = domain.zero
            for i in range(self.row_dim):
                a = self.block(i, j)
                x = domain.axpy(1, a.adjoint(ys[i]), x)
            xs.append(x)
        return xs


class ColumnLinearOperator(LinearOperator, BlockStructure):
    """
    Class for linear operators acting between a space and a direct sum of spaces.
    """

    def __init__(self, operators):
        """
        Args:
            operators ([LinearOperator]): List of operators all with a common domain.
        """

        self._operators = operators

        domains = [operator.domain for operator in operators]
        codomains = [operator.codomain for operator in operators]

        domain = domains[0]
        assert all(operator.domain == domain for operator in operators)

        codomain = HilbertSpaceDirectSum(codomains)

        blocks = [[operator] for operator in operators]

        block_operator = BlockLinearOperator(blocks)

        def mapping(x):
            return block_operator([x])

        def adjoint_mapping(ys):
            return block_operator.adjoint(ys)[0]

        super().__init__(domain, codomain, mapping, adjoint_mapping=adjoint_mapping)

        row_dim = len(self._operators)
        BlockStructure.__init__(self, row_dim, 1)

    def block(self, i, j):
        """
        Returns the operator in the (i,j)th sub-block.
        """
        assert 0 <= i < self.row_dim
        assert j == 0

        return self._operators[i]


class RowLinearOperator(LinearOperator, BlockStructure):
    """
    Class for linear operators acting between a direct sum of spaces and a space.
    """

    def __init__(self, operators):
        """
        Args:
            operators ([LinearOperator]): List of operators all with a common codomain.
        """

        self._operators = operators

        domains = [operator.domain for operator in operators]
        codomains = [operator.codomain for operator in operators]

        codomain = codomains[0]
        assert all(operator.codomain == codomain for operator in operators)

        domain = HilbertSpaceDirectSum(domains)

        blocks = [operators]

        block_operator = BlockLinearOperator(blocks)

        def mapping(xs):
            return block_operator(xs)[0]

        def adjoint_mapping(y):
            return block_operator.adjoint([y])

        super().__init__(domain, codomain, mapping, adjoint_mapping=adjoint_mapping)

        col_dim = len(self._operators)
        BlockStructure.__init__(self, 1, col_dim)

    def block(self, i, j):
        """
        Returns the operator in the (i,j)th sub-block.
        """
        assert i == 0
        assert 0 <= j < self.col_dim

        return self._operators[j]


class BlockDiagonalLinearOperator(LinearOperator, BlockStructure):
    """
    Class for linear operators acting between direct sums of Hilbert spaces that are defined
    in a blockwise diagonal manner.

    Instances are formed from list of LinearOperators that define the diagonal blocks.
    """

    def __init__(self, operators):
        """
        Args:
            operators ([LinearOperator]): List of operators for the diagonal blocks.
        """
        self._operators = operators

        domain = HilbertSpaceDirectSum([operator.domain for operator in operators])
        codomain = HilbertSpaceDirectSum([operator.codomain for operator in operators])

        def mapping(xs):
            return [operator(x) for operator, x in zip(operators, xs)]

        def adjoint_mapping(ys):
            return [operator.adjoint(y) for operator, y in zip(operators, ys)]

        super().__init__(domain, codomain, mapping, adjoint_mapping=adjoint_mapping)

        row_dim = len(self._operators)
        col_dim = row_dim
        BlockStructure.__init__(self, row_dim, col_dim)

    def block(self, i, j):
        """
        Returns the operator in the (i,j)th sub-block.
        """
        assert 0 <= i < self.row_dim
        assert 0 <= j < self.col_dim

        if i == j:
            return self._operators[i]
        else:
            domain = self._operators[j].domain
            codomain = self._operators[i].codomain
            return domain.zero_operator(codomain)



"""
This module contains functions used to check instances of the classed defined within hilbert.py
"""

import numpy as np


class HilbertSpaceChecks:
    """
    Class defining consistency checks for a HilbertSpace.
    """

    def __init__(self, space, /, *, rtol=1.0e-6, atol=1.0e-6):
        self._space = space
        self._rtol = rtol
        self._atol = atol

    @property
    def space(self):
        """
        Return the space
        """
        return self._space

    @property
    def rtol(self):
        """
        Return the relative tolerance for numerical tests.
        """
        return self._rtol

    @property
    def atol(self):
        """
        Return the absolute tolerance for numerical tests.
        """
        return self._atol

    def _list_of_checks(self):
        return [
            attribute
            for attribute in dir(self)
            if callable(getattr(self, attribute)) and attribute.startswith("check")
        ]

    def passed_checks(self):
        """
        Return a list of checks that have passed.
        """
        results_and_checks = [
            [getattr(self, check)(), check] for check in self._list_of_checks()
        ]
        return [check for value, check in results_and_checks if value]

    def failed_checks(self):
        """
        Return a list of checks that have failed.
        """
        results_and_checks = [
            [getattr(self, check)(), check] for check in self._list_of_checks()
        ]
        return [check for value, check in results_and_checks if not value]

    def all_checks_passed(self, /, *, trials=1):
        """
        True if all tests have passed. By default one random trial
        of the tests is performed. Optionally, multiple random trials
        can be run by setting the trials variable > 1.
        """
        okay = True
        for _ in range(trials):
            okay = okay and len(self.failed_checks()) == 0
        return okay

    def check_norm_non_negativity(self):
        """
        True if the norm returns non-negative values.
        """
        x = self.space.random()
        return self.space.norm(x) >= 0

    def check_norm_homogeneity(self):
        """
        True if the norm is positively homogeneous of order 1.
        """
        a = np.random.randn()
        x1 = self.space.random()
        x2 = self.space.multiply(a, x1)
        norm1 = np.abs(a) * self.space.norm(x1)
        norm2 = self.space.norm(x2)
        diff = np.abs(norm2 - norm1) / norm1 if norm1 > 0 else 0
        return diff < self.rtol

    def check_triangle_inequality(self):
        """
        True if triangle inequality is satisfied.
        """
        x1 = self.space.random()
        x2 = self.space.random()
        x3 = self.space.add(x1, x2)
        norm1 = self.space.norm(x1)
        norm2 = self.space.norm(x2)
        norm3 = self.space.norm(x3)
        diff = (norm1 + norm2 - norm3) / (norm1 + norm2)
        return diff > -self.rtol

    def check_inner_product_symmetry(self):
        """
        True if the inner product is symmetric in its arguments.
        """
        x1 = self.space.random()
        x2 = self.space.random()
        ip1 = self.space.inner_product(x1, x2)
        ip2 = self.space.inner_product(x2, x1)
        diff = np.abs(ip2 - ip1) / np.abs(ip1)
        return diff < self.rtol

    def check_inner_product_bilinearity(self):
        """
        True if the inner product is bilinear. Based on check of first
        argument only, assumes the symmetry condition has been met.
        """
        a1 = np.random.randn()
        a2 = np.random.randn()
        x1 = self.space.random()
        x2 = self.space.random()
        x3 = self.space.random()
        x4 = self.space.add(self.space.multiply(a1, x1), self.space.multiply(a2, x2))
        ip1 = a1 * self.space.inner_product(x1, x3) + a2 * self.space.inner_product(
            x2, x3
        )
        ip2 = self.space.inner_product(x4, x3)
        diff = np.abs(ip2 - ip1) / np.abs(ip1)
        return diff < self.rtol

    def check_riesz_mapping_definition(self):
        """
        Checks isomorphism from dual to space is valid.
        """
        x1 = self.space.random()
        xp = self.space.dual.random()
        p1 = xp(x1)
        x2 = self.space.from_dual(xp)
        p2 = self.space.inner_product(x2, x1)
        diff = np.abs(p2 - p1) / np.abs(p1)
        return diff < self.rtol

    def check_inverse_riesz_mapping_definition(self):
        """
        True if riesz mapping and its inverse are compatible.
        """
        x1 = self.space.random()
        xp = self.space.to_dual(x1)
        x2 = self.space.from_dual(xp)
        x3 = self.space.subtract(x1, x2)
        diff = self.space.norm(x3) / self.space.norm(x1)
        return diff < self.rtol

    def check_riesz_mapping_linearity(self):
        """
        True if the riesz mapping is linear.
        """
        xp1 = self.space.dual.random()
        xp2 = self.space.dual.random()
        x1 = self.space.from_dual(xp1)
        x2 = self.space.from_dual(xp2)
        xp3 = xp1 + xp2
        x3 = self.space.from_dual(xp3)
        x4 = self.space.subtract(x3, self.space.add(x1, x2))
        diff = self.space.norm(x4) / self.space.norm(x3)
        return diff < self.rtol

    def check_inverse_riesz_mapping_linearity(self):
        """
        True if the riesz mapping is linear.
        """
        x1 = self.space.random()
        x2 = self.space.random()
        xp1 = self.space.to_dual(x1)
        xp2 = self.space.to_dual(x2)
        x3 = self.space.add(x1, x2)
        xp3 = self.space.to_dual(x3)
        xp4 = xp3 - xp1 - xp2
        diff = self.space.dual.norm(xp4) / self.space.dual.norm(xp3)
        return diff < self.rtol

    def check_to_component_linearity(self):
        """
        True if to_component mapping is linear.
        """
        a1 = np.random.randn()
        a2 = np.random.randn()
        x1 = self.space.random()
        x2 = self.space.random()
        x3 = self.space.add(self.space.multiply(a1, x1), self.space.multiply(a2, x2))
        c1 = self.space.to_components(x1)
        c2 = self.space.to_components(x2)
        c3 = self.space.to_components(x3)
        c4 = c3 - a1 * c1 - a2 * c2
        diff = np.linalg.norm(c4) / np.linalg.norm(c1)
        return diff < self.rtol

    def check_from_component_linearity(self):
        """
        True if from_component mapping is linear.
        """
        dim = self.space.dim
        a1 = np.random.randn()
        a2 = np.random.randn()
        c1 = np.random.randn(dim)
        c2 = np.random.randn(dim)
        c3 = a1 * c1 + a2 * c2
        x1 = self.space.from_components(c1)
        x2 = self.space.from_components(c2)
        x3 = self.space.from_components(c3)
        x4 = self.space.subtract(
            x3, self.space.add(self.space.multiply(a1, x1), self.space.multiply(a2, x2))
        )
        diff = self.space.norm(x4) / self.space.norm(x3)
        return diff < self.rtol

    def check_componet_mapping_compatibility(self):
        """
        True if to_component and from_component mappings are
        mutual inverses.
        """
        x1 = self.space.random()
        c1 = self.space.to_components(x1)
        x2 = self.space.from_components(c1)
        x3 = self.space.subtract(x1, x2)
        diff = self.space.norm(x3) / self.space.norm(x1)
        return diff < self.rtol



"""
Module for linear solvers. 
"""

from abc import ABC, abstractmethod
import numpy as np
from scipy.sparse.linalg import LinearOperator as ScipyLinOp
from scipy.linalg import (
    cho_factor,
    cho_solve,
    lu_factor,
    lu_solve,
)
from scipy.sparse.linalg import gmres, bicgstab, cg, bicg
from pygeoinf.hilbert_space import LinearOperator


class LinearSolver(ABC):
    """
    Abstract base class for linear solvers.
    """


class DirectLinearSolver(LinearSolver):
    """
    Abstract base class for direct linear solvers.
    """


class LUSolver(DirectLinearSolver):
    """
    Direct Linear solver class based on LU decomposition of the
    matrix representation.
    """

    def __init__(self, /, *, galerkin=False):
        """
        Args:
            galerkin (bool): If true, the Galerkin matrix representation is used.
        """
        self._galerkin = galerkin

    def __call__(self, operator):
        """
        Returns the inverse of a LinearOperator based on the LU factorisation
        of its dense matrix representation.
        """

        assert operator.domain.dim == operator.codomain.dim
        matrix = operator.matrix(dense=True, galerkin=self._galerkin)
        factor = lu_factor(matrix, overwrite_a=True)

        def matvec(cy):
            return lu_solve(factor, cy, 0)

        def rmatvec(cx):
            return lu_solve(factor, cx, 1)

        inverse_matrix = ScipyLinOp(
            (operator.domain.dim, operator.codomain.dim),
            matvec=matvec,
            rmatvec=rmatvec,
        )

        return LinearOperator.from_matrix(
            operator.codomain, operator.domain, inverse_matrix, galerkin=self._galerkin
        )


class CholeskySolver(DirectLinearSolver):
    """
    Direct Linear solver class based on Cholesky decomposition of the
    matrix representation. It is assumed that the operator's matrix
    representation is self-adjoint and positive-definite.
    """

    def __init__(self, /, *, galerkin=False):
        """
        galerkin (bool): If true, use the Galerkin matrix representation.
        """
        self._galerkin = galerkin

    def __call__(self, operator):
        """
        Returns the inverse of a LinearOperator based on the LU factorisation
        of its dense matrix representation.
        """

        assert operator.is_automorphism

        matrix = operator.matrix(dense=True, galerkin=self._galerkin)
        factor = cho_factor(matrix, overwrite_a=False)

        def matvec(cy):
            return cho_solve(factor, cy)

        inverse_matrix = ScipyLinOp(
            (operator.domain.dim, operator.codomain.dim), matvec=matvec, rmatvec=matvec
        )

        return LinearOperator.from_matrix(
            operator.domain, operator.domain, inverse_matrix, galerkin=self._galerkin
        )


class IterativeLinearSolver(LinearSolver):
    """
    Abstract base class for direct linear solvers.
    """

    @abstractmethod
    def solve_linear_system(self, operator, preconditioner, y, x0):
        """
        Returns the solution of the linear system.
        """

    def solve_adjoint_linear_system(self, operator, preconditioner, x, y0):
        """
        Returns the solution of the adjoint linear system.
        """
        return self.solve_linear_system(operator.adjoint, preconditioner, x, y0)

    def __call__(self, operator, /, *, preconditioner=None):
        assert operator.is_automorphism
        return LinearOperator(
            operator.codomain,
            operator.domain,
            lambda y: self.solve_linear_system(operator, preconditioner, y, None),
            adjoint_mapping=lambda x: self.solve_adjoint_linear_system(
                operator, preconditioner, x, None
            ),
        )


class CGMatrixSolver(IterativeLinearSolver):
    """
    Linear solver for self-adjoint operators based on the application
    of the conjugate gradient algorithm to the matrix representation.

    It is assumed that the matrix representation of the operator
    is also self-adjoint. This will hold automatically when the
    Galerkin representation is used.
    """

    def __init__(
        self, /, *, galerkin=False, rtol=1.0e-5, atol=0, maxiter=None, callback=None
    ):
        """
        Args:
            galerkin (bool): True if the Galerkin matrix representation is used.
            rtol (float): relative tolerance within convergence checks.
            atol (float): absolute tolerance within convergence checks.
            maxiter (int): maximum number of iterations to allow.
            callback (callable): callable function after each iteration. This function
                takes in as argument the current solution vector.
        """
        self._galerkin = galerkin
        self._rtol = rtol
        self._atol = atol
        self._maxiter = maxiter
        self._callback = callback

    def solve_linear_system(self, operator, preconditioner, y, x0):

        domain = operator.codomain
        matrix = operator.matrix(galerkin=self._galerkin)

        if preconditioner is None:
            matrix_preconditioner = None
        else:
            matrix_preconditioner = preconditioner.matrix(galerkin=self._galerkin)

        cx0 = None if x0 is None else domain.to_components(x0)
        cy = domain.to_components(y)

        cxp = cg(
            matrix,
            cy,
            x0=cx0,
            rtol=self._rtol,
            atol=self._atol,
            maxiter=self._maxiter,
            M=matrix_preconditioner,
            callback=self._callback,
        )[0]
        if self._galerkin:
            xp = domain.dual.from_components(cxp)
            return domain.from_dual(xp)
        else:
            return domain.from_components(cxp)


class BICGMatrixSolver(IterativeLinearSolver):
    """
    Linear solver for general operators based on the application
    of the biconjugate gradient algorithm to the matrix representation.
    """

    def __init__(
        self, /, *, galerkin=False, rtol=1.0e-5, atol=0, maxiter=None, callback=None
    ):
        """
        Args:
            galerkin (bool): True if the Galerkin matrix representation is used.
            rtol (float): relative tolerance within convergence checks.
            atol (float): absolute tolerance within convergence checks.
            maxiter (int): maximum number of iterations to allow.
            callback (callable): callable function after each iteration. This function
                takes in as argument the current solution vector.
        """
        self._galerkin = galerkin
        self._rtol = rtol
        self._atol = atol
        self._maxiter = maxiter
        self._callback = callback

    def solve_linear_system(self, operator, preconditioner, y, x0):

        domain = operator.codomain
        codomain = operator.domain
        matrix = operator.matrix(galerkin=self._galerkin)

        if preconditioner is None:
            matrix_preconditioner = None
        else:
            matrix_preconditioner = preconditioner.matrix(galerkin=self._galerkin)

        cx0 = None if x0 is None else domain.to_components(x0)
        cy = domain.to_components(y)

        cxp = bicg(
            matrix,
            cy,
            x0=cx0,
            rtol=self._rtol,
            atol=self._atol,
            maxiter=self._maxiter,
            M=matrix_preconditioner,
            callback=self._callback,
        )[0]
        if self._galerkin:
            xp = codomain.dual.from_components(cxp)
            return codomain.from_dual(xp)
        else:
            return codomain.from_components(cxp)


class BICGStabMatrixSolver(IterativeLinearSolver):
    """
    Linear solver for general operators based on the application
    of the biconjugate gradient stabilised algorithm to the matrix representation.
    """

    def __init__(
        self, /, *, galerkin=False, rtol=1.0e-5, atol=0, maxiter=None, callback=None
    ):
        """
        Args:
            galerkin (bool): True if the Galerkin matrix representation is used.
            rtol (float): relative tolerance within convergence checks.
            atol (float): absolute tolerance within convergence checks.
            maxiter (int): maximum number of iterations to allow.
            callback (callable): callable function after each iteration. This function
                takes in as argument the current solution vector.
        """
        self._galerkin = galerkin
        self._rtol = rtol
        self._atol = atol
        self._maxiter = maxiter
        self._callback = callback

    def solve_linear_system(self, operator, preconditioner, y, x0):

        domain = operator.codomain
        codomain = operator.domain
        matrix = operator.matrix(galerkin=self._galerkin)

        if preconditioner is None:
            matrix_preconditioner = None
        else:
            matrix_preconditioner = preconditioner.matrix(galerkin=self._galerkin)

        cx0 = None if x0 is None else domain.to_components(x0)
        cy = domain.to_components(y)

        cxp = bicgstab(
            matrix,
            cy,
            x0=cx0,
            rtol=self._rtol,
            atol=self._atol,
            maxiter=self._maxiter,
            M=matrix_preconditioner,
            callback=self._callback,
        )[0]
        if self._galerkin:
            xp = codomain.dual.from_components(cxp)
            return codomain.from_dual(xp)
        else:
            return codomain.from_components(cxp)


class GMRESMatrixSolver(IterativeLinearSolver):
    """
    Linear solver for general operators based on the application
    of the GMRES algorithm to the matrix representation.
    """

    def __init__(
        self,
        /,
        *,
        galerkin=False,
        rtol=1.0e-5,
        atol=0,
        restart=None,
        maxiter=None,
        callback=None,
        callback_type=None,
    ):
        """
        Args:
            galerkin (bool): True if the Galerkin matrix representation is used.
            rtol (float): relative tolerance within convergence checks.
            atol (float): absolute tolerance within convergence checks.
            restart (int): Number of iterations between restarts.
            maxiter (int): maximum number of iterations to allow.
            callback (callable): callable function after each iteration. Signature
                of this function is determined by callback_type.
            callback_type ("x", "pr_norm", "legacy"): If "x" the current solution is
                passed to the callback function, if "pr_norm" it is the preconditioned
                residual norm. The default is "legacy" which means the same as "pr_norm",
                but changes the meaning of maxiter to count inner iterations instead of
                restart cycles.
        """
        self._galerkin = galerkin
        self._rtol = rtol
        self._atol = atol
        self._restart = restart
        self._maxiter = maxiter
        self._callback = callback
        self._callback_type = callback_type

    def solve_linear_system(self, operator, preconditioner, y, x0):

        domain = operator.codomain
        codomain = operator.domain
        matrix = operator.matrix(galerkin=self._galerkin)

        if preconditioner is None:
            matrix_preconditioner = None
        else:
            matrix_preconditioner = preconditioner.matrix(galerkin=self._galerkin)

        cx0 = None if x0 is None else domain.to_components(x0)
        cy = domain.to_components(y)

        cxp = gmres(
            matrix,
            cy,
            x0=cx0,
            rtol=self._rtol,
            atol=self._atol,
            restart=self._restart,
            maxiter=self._maxiter,
            M=matrix_preconditioner,
            callback=self._callback,
            callback_type=self._callback_type,
        )[0]

        if self._galerkin:
            xp = codomain.dual.from_components(cxp)
            return codomain.from_dual(xp)
        else:
            return codomain.from_components(cxp)


class CGSolver(IterativeLinearSolver):
    """
    LinearSolver class using the conjugate gradient algorithm without
    use of the matrix representation. Can be applied to self-adjoint
    operators on a general Hilbert space.
    """

    def __init__(self, /, *, rtol=1.0e-5, atol=0, maxiter=None, callback=None):
        """
        Args:
            rtol (float): relative tolerance within convergence checks.
            atol (float): absolute tolerance within convergence checks.
            maxiter (int): maximum number of iterations to allow.
            callback (callable): callable function after each iteration. This function
                takes in as argument the current solution vector.
        """
        if rtol > 0:
            self._rtol = rtol
        else:
            raise ValueError("rtol must be positive")
        if atol >= 0:
            self._atol = atol
        else:
            raise ValueError("atol must be non-negative!")
        if maxiter is None:
            self._maxiter = maxiter
        else:
            if maxiter >= 0:
                self._maxiter = maxiter
            else:
                raise ValueError("maxiter must be None or positive")

        self._callback = callback

    def solve_linear_system(self, operator, preconditioner, y, x0):

        domain = operator.domain
        if x0 is None:
            x = domain.zero
        else:
            x = domain.copy(x0)

        r = domain.subtract(y, operator(x))
        if preconditioner is None:
            z = domain.copy(r)
        else:
            z = preconditioner(r)
        p = domain.copy(z)

        y_squared_norm = domain.squared_norm(y)
        if y_squared_norm <= self._atol:
            return y

        tol = np.max([self._atol, self._rtol * y_squared_norm])

        if self._maxiter is None:
            maxiter = 10 * domain.dim
        else:
            maxiter = self._maxiter

        for _ in range(maxiter):

            if domain.norm(r) <= tol:
                break

            q = operator(p)
            num = domain.inner_product(r, z)
            den = domain.inner_product(p, q)
            alpha = num / den

            x = domain.axpy(alpha, p, x)
            r = domain.axpy(-alpha, q, r)

            if preconditioner is None:
                z = domain.copy(r)
            else:
                z = preconditioner(r)

            den = num
            num = operator.domain.inner_product(r, z)
            beta = num / den

            p = domain.multiply(beta, p)
            p = domain.add(p, z)

            if self._callback is not None:
                self._callback(x)

        return x



"""
Module for Sobolev spaces on the two-sphere.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import diags, coo_array
import pyshtools as sh

import matplotlib.ticker as mticker
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter


from pygeoinf.hilbert_space import LinearOperator, EuclideanSpace
from pygeoinf.homogeneous_space.homogeneous_space import HomogeneousSpaceSobolev
from pygeoinf.gaussian_measure import GaussianMeasure


class Sobolev(HomogeneousSpaceSobolev):
    """Sobolev spaces on a two-sphere as an instance of HilbertSpace."""

    def __init__(
        self, lmax, order, scale, /, *, vector_as_SHGrid=True, radius=1, grid="DH"
    ):
        """
        Args:
            lmax (int): Truncation degree for spherical harmoincs.
            order (float): Order of the Sobolev space.
            scale (float): Non-dimensional length-scale for the space.
            vector_as_SHGrid (bool): If true, elements of the space are
                instances of the SHGrid class, otherwise they are SHCoeffs.
            radius (float): Radius of the two sphere.
            grid (str): pyshtools grid type.
        """

        self._lmax = lmax
        self._radius = radius
        self._grid = grid
        if self.grid == "DH2":
            self._sampling = 2
        else:
            self._sampling = 1
        self._extend = True
        self._normalization = "ortho"
        self._csphase = 1
        self._sparse_coeffs_to_component = self._coefficient_to_component_mapping()

        dim = (lmax + 1) ** 2

        self._vector_as_SHGrid = vector_as_SHGrid
        if vector_as_SHGrid:
            HomogeneousSpaceSobolev.__init__(
                self,
                order,
                scale,
                dim,
                self._to_components_from_SHGrid,
                self._from_components_to_SHGrid,
                self._inner_product_impl,
                self._to_dual_impl,
                self._from_dual_impl,
                vector_multiply=self._vector_multiply_impl,
            )
        else:
            HomogeneousSpaceSobolev.__init__(
                self,
                order,
                scale,
                dim,
                self._to_components_from_SHCoeffs,
                self._from_components_to_SHCoeffs,
                self._inner_product_impl,
                self._to_dual_impl,
                self._from_dual_impl,
                vector_multiply=self._vector_multiply_impl,
            )

        self._metric_tensor = self._degree_dependent_scaling_to_diagonal_matrix(
            self._sobolev_function
        )
        self._inverse_metric_tensor = self._degree_dependent_scaling_to_diagonal_matrix(
            lambda l: 1 / self._sobolev_function(l)
        )

    # ===============================================#
    #                   Properties                  #
    # ===============================================#

    @property
    def lmax(self):
        """Returns truncation degree."""
        return self._lmax

    @property
    def radius(self):
        """Returns radius of the sphere."""
        return self._radius

    @property
    def grid(self):
        """Returns spatial grid option."""
        return self._grid

    @property
    def extend(self):
        """True is spatial grid contains longitudes 0 and 360."""
        return self._extend

    @property
    def normalization(self):
        """Spherical harmonic normalisation option."""
        return self._normalization

    @property
    def csphase(self):
        """Condon Shortley phase option."""
        return self._csphase

    # ==============================================#
    #                 Public methods                #
    # ==============================================#

    def random_point(self):
        latitude = np.random.uniform(-90, 90)
        longitude = np.random.uniform(0, 360)
        return [latitude, longitude]

    def low_degree_projection(self, truncation_degree, /, *, smoother=None):
        """
        Returns a LinearOperator that maps the space onto a Sobolev space with
        the same parameters but based on a lower truncation degree.
        """
        truncation_degree = (
            truncation_degree if truncation_degree <= self.lmax else self.lmax
        )
        f = smoother if smoother is not None else lambda l: l

        # construct the spare matrix that performs the coordinate projection.
        row_dim = (truncation_degree + 1) ** 2
        col_dim = (self.lmax + 1) ** 2

        row = 0
        col = 0
        rows = []
        cols = []
        data = []
        for l in range(self.lmax + 1):
            fac = f(l)
            for _ in range(l + 1):
                if l <= truncation_degree:
                    rows.append(row)
                    row += 1
                    cols.append(col)
                    data.append(fac)
                col += 1

        for l in range(truncation_degree + 1):
            fac = f(l)
            for _ in range(1, l + 1):
                rows.append(row)
                row += 1
                cols.append(col)
                data.append(fac)
                col += 1

        smat = coo_array(
            (data, (rows, cols)), shape=(row_dim, col_dim), dtype=int
        ).tocsc()

        codomain = Sobolev(
            truncation_degree,
            self.order,
            self.scale,
            vector_as_SHGrid=self._vector_as_SHGrid,
            radius=self.radius,
            grid=self._grid,
        )

        def mapping(u):
            uc = self.to_components(u)
            vc = smat @ uc
            return codomain.from_components(vc)

        def adjoint_mapping(v):
            vc = codomain.to_components(v)
            uc = smat.T @ vc
            return self.from_components(uc)

        return LinearOperator(self, codomain, mapping, adjoint_mapping=adjoint_mapping)

    def dirac(self, point):

        latitude, longitude = point
        colatitude = 90 - latitude

        coeffs = sh.expand.spharm(
            self.lmax,
            colatitude,
            longitude,
            normalization="ortho",
            degrees=True,
        )
        c = self._to_components_from_coeffs(coeffs)
        return self.dual.from_components(c)

    def invariant_automorphism(self, f):
        matrix = self._degree_dependent_scaling_to_diagonal_matrix(f)

        def mapping(x):
            return self.from_components(matrix @ self.to_components(x))

        return LinearOperator.self_adjoint(self, mapping)

    def invariant_gaussian_measure(self, f, /, *, expectation=None):

        def g(l):
            return np.sqrt(f(l) / (self.radius**2 * self._sobolev_function(l)))

        def h(l):
            return np.sqrt(self.radius**2 * self._sobolev_function(l) * f(l))

        matrix = self._degree_dependent_scaling_to_diagonal_matrix(g)
        adjoint_matrix = self._degree_dependent_scaling_to_diagonal_matrix(h)
        domain = EuclideanSpace(self.dim)

        def mapping(c):
            return self.from_components(matrix @ c)

        def adjoint_mapping(u):
            return adjoint_matrix @ self.to_components(u)

        inverse_matrix = self._degree_dependent_scaling_to_diagonal_matrix(
            lambda l: 1 / g(l)
        )

        inverse_adjoint_matrix = self._degree_dependent_scaling_to_diagonal_matrix(
            lambda l: 1 / h(l)
        )

        def inverse_mapping(u):
            return inverse_matrix @ self.to_components(u)

        def inverse_adjoint_mapping(c):
            return self.from_components(inverse_adjoint_matrix @ c)

        covariance_factor = LinearOperator(
            domain, self, mapping, adjoint_mapping=adjoint_mapping
        )

        inverse_covariance_factor = LinearOperator(
            self, domain, inverse_mapping, adjoint_mapping=inverse_adjoint_mapping
        )

        return GaussianMeasure(
            covariance_factor=covariance_factor,
            inverse_covariance_factor=inverse_covariance_factor,
            expectation=expectation,
        )

    def plot(
        self,
        f,
        /,
        *,
        projection=ccrs.PlateCarree(),
        contour=False,
        cmap="RdBu",
        coasts=False,
        rivers=False,
        borders=False,
        map_extent=None,
        gridlines=True,
        symmetric=False,
        **kwargs,
    ):
        """
        Return a plot of a function.

        Args:
            f (vector): Function to be plotted.
            projection: cartopy projection to be used. Default is Robinson.
            contour (bool): If True, a contour plot is made, otherwise a pcolor plot.
            cmap (string): colormap. Default is RdBu.
            coasts (bool): If True, coast lines plotted. Default is True.
            rivers (bool): If True, major rivers plotted. Default is False.
            borders (bool): If True, country borders are plotted. Default is False.
            map_extent ([float]): Sets the (lon, lat) range for plotting.
            Tuple of [lon_min, lon_max, lat_min, lat_max]. Default is None.
            gridlines (bool): If True, gridlines are included. Default is True.
            symmetric (bool): If True, clim values set symmetrically based on the fields maximum absolute value.
                Option overridden if vmin or vmax are set.
            kwargs: Keyword arguments for forwarding to the plotting functions.
        """

        if self._vector_as_SHGrid:
            field = f
        else:
            field = f.expand(normalization=self.normalization, csphase=self.csphase)

        lons = field.lons()
        lats = field.lats()

        figsize = kwargs.pop("figsize", (10, 8))
        fig, ax = plt.subplots(figsize=figsize, subplot_kw={"projection": projection})

        if map_extent is not None:
            ax.set_extent(map_extent, crs=ccrs.PlateCarree())

        if coasts:
            ax.add_feature(cfeature.COASTLINE, linewidth=0.8)

        if rivers:
            ax.add_feature(cfeature.RIVERS, linewidth=0.8)

        if borders:
            ax.add_feature(cfeature.BORDERS, linewidth=0.8)

        kwargs.setdefault("cmap", cmap)

        lat_interval = kwargs.pop("lat_interval", 30)
        lon_interval = kwargs.pop("lon_interval", 30)

        if symmetric:
            data_max = 1.2 * np.nanmax(np.abs(f.data))
            kwargs.setdefault("vmin", -data_max)
            kwargs.setdefault("vmax", data_max)

        levels = kwargs.pop("levels", 10)

        if contour:

            im = ax.contourf(
                lons,
                lats,
                field.data,
                transform=ccrs.PlateCarree(),
                levels=levels,
                **kwargs,
            )

        else:

            im = ax.pcolormesh(
                lons,
                lats,
                field.data,
                transform=ccrs.PlateCarree(),
                **kwargs,
            )

        if gridlines:
            gl = ax.gridlines(
                linestyle="--",
                draw_labels=True,
                dms=True,
                x_inline=False,
                y_inline=False,
            )

            gl.xlocator = mticker.MultipleLocator(lon_interval)
            gl.ylocator = mticker.MultipleLocator(lat_interval)
            gl.xformatter = LongitudeFormatter()
            gl.yformatter = LatitudeFormatter()

        return fig, ax, im

    # ==============================================#
    #                Private methods                #
    # ==============================================#

    def _coefficient_to_component_mapping(self):
        # returns a sparse matrix that maps from flattended pyshtools
        # coefficients to component vectors.

        row_dim = (self.lmax + 1) ** 2
        col_dim = 2 * (self.lmax + 1) ** 2

        row = 0
        rows = []
        cols = []
        for l in range(self.lmax + 1):
            col = l * (self.lmax + 1)
            for _ in range(l + 1):
                rows.append(row)
                row += 1
                cols.append(col)
                col += 1

        for l in range(self.lmax + 1):
            col = (self.lmax + 1) ** 2 + l * (self.lmax + 1) + 1
            for _ in range(1, l + 1):
                rows.append(row)
                row += 1
                cols.append(col)
                col += 1

        data = [1] * row_dim
        return coo_array(
            (data, (rows, cols)), shape=(row_dim, col_dim), dtype=int
        ).tocsc()

    def spherical_harmonic_index(self, l, m):
        """Return the component index for given spherical harmonic degree and order."""
        if m >= 0:
            return int(l * (l + 1) / 2) + m
        else:
            offset = int((self.lmax + 1) * (self.lmax + 2) / 2)
            return offset + int((l - 1) * l / 2) - m - 1

    def _to_components_from_coeffs(self, coeffs):
        """Return component vector from coefficient array."""
        f = coeffs.flatten(order="C")
        return self._sparse_coeffs_to_component @ f

    def _to_components_from_SHCoeffs(self, ulm):
        """Return component vector from SHCoeffs object."""
        return self._to_components_from_coeffs(ulm.coeffs)

    def _to_components_from_SHGrid(self, u):
        """Return component vector from SHGrid object."""
        ulm = u.expand(normalization=self.normalization, csphase=self.csphase)
        return self._to_components_from_SHCoeffs(ulm)

    def _from_components_to_SHCoeffs(self, c):
        """Return SHCoeffs object from its component vector."""
        f = self._sparse_coeffs_to_component.T @ c
        coeffs = f.reshape((2, self.lmax + 1, self.lmax + 1))
        return sh.SHCoeffs.from_array(
            coeffs, normalization=self.normalization, csphase=self.csphase
        )

    def _from_components_to_SHGrid(self, c):
        """Return SHGrid object from its component vector."""
        ulm = self._from_components_to_SHCoeffs(c)
        return ulm.expand(grid=self.grid, extend=self.extend)

    def _degree_dependent_scaling_to_diagonal_matrix(self, f):
        values = np.zeros(self.dim)
        i = 0
        for l in range(self.lmax + 1):
            j = i + l + 1
            values[i:j] = f(l)
            i = j
        for l in range(1, self.lmax + 1):
            j = i + l
            values[i:j] = f(l)
            i = j
        return diags([values], [0])

    def _sobolev_function(self, l):
        # Degree-dependent scaling that defines the Sobolev inner product.
        return (1 + self.scale**2 * l * (l + 1)) ** self.order

    def _inner_product_impl(self, u, v):
        # Implementation of the inner product.
        return self.radius**2 * np.dot(
            self._metric_tensor @ self.to_components(u), self.to_components(v)
        )

    def _to_dual_impl(self, u):
        # Implementation of the mapping to the dual space.
        c = self._metric_tensor @ self.to_components(u) * self.radius**2
        return self.dual.from_components(c)

    def _from_dual_impl(self, up):
        # Implementation of the mapping from the dual space.
        c = self._inverse_metric_tensor @ self.dual.to_components(up) / self.radius**2
        return self.from_components(c)

    def _normalise_covariance_function(self, f, amplitude):
        # Normalise a degree-dependent scaling function, f, so that
        # the associated invariant Gaussian measure has standard deviation
        # for point values equal to amplitude.
        norm = 0
        for l in range(self.lmax + 1):
            norm += (
                f(l)
                * (2 * l + 1)
                / (4 * np.pi * self.radius**2 * self._sobolev_function(l))
            )
        return lambda l: amplitude**2 * f(l) / norm

    def _vector_multiply_impl(self, u1, u2):

        if self._vector_as_SHGrid:
            return u1 * u2
        else:
            u1_field = u1.expand(grid=self.grid, extend=self.extend)
            u2_field = u2.expand(grid=self.grid, extend=self.extend)
            u3_field = u1_field * u2_field
            return u3_field.expand(
                normalization=self.normalization, csphase=self.csphase
            )


class Lebesgue(Sobolev):
    """
    L2 on the two-sphere as an instance of HilbertSpace.

    Implemented as a special case of the Sobolev class with order = 0.
    """

    def __init__(self, lmax, /, *, vector_as_SHGrid=True, radius=1, grid="DH"):
        """
        Args:
            lmax (int): Truncation degree for spherical harmoincs.
            vector_as_SHGrid (bool): If true, elements of the space are
                instances of the SHGrid class, otherwise they are SHCoeffs.
            radius (float): Radius of the two sphere.
            grid (str): pyshtools grid type.
        """
        super().__init__(
            lmax,
            0,
            1,
            vector_as_SHGrid=vector_as_SHGrid,
            radius=radius,
            grid=grid,
        )


###############################################################
#                      Utility classes                        #
###############################################################


class LowPassFilter:
    """
    Class implementing a simple Hann-type low-pass filter in
    the spherical harmonic domain
    """

    def __init__(self, lower_degree, upper_degree):
        """
        Args:
            lower_degree (int): Below this degree, the filter returns one.
            upper_degree (int): Above this degree, the filter returns zero.
                Its value between the lower and upper degrees decreases smoothly.
        """
        self._lower_degree = lower_degree
        self._upper_degree = upper_degree

    def __call__(self, l):
        if l <= self._lower_degree:
            return 1
        elif self._lower_degree <= l <= self._upper_degree:
            return 0.5 * (
                1
                - np.cos(
                    np.pi
                    * (self._upper_degree - l)
                    / (self._upper_degree - self._lower_degree)
                )
            )
        else:
            return 0
